/////////////////////////////////////////////////////////////////////////////////
//+++ do not change
/////////////////////////////////////////////////////////////////////////////////
#if defined(_WIN64) || defined(_WIN32) //MSVC Compiler
#define MY_EXPORT __declspec(dllexport)
#else
#define MY_EXPORT
#endif

/////////////////////////////////////////////////////////////////////////////////
//+++ h-files
/////////////////////////////////////////////////////////////////////////////////
#include <math.h>
#include <iostream>
#include <gsl/gsl_vector.h>
#include <gsl/gsl_matrix.h>
#include <gsl/gsl_math.h>
std::string fitFunctionPath="Z:/qtiSAS/FitFunctions/";
std::string OS="WINDOWS";

/////////////////////////////////////////////////////////////////////////////////
//+++ Parameters Structure
/////////////////////////////////////////////////////////////////////////////////
#ifndef  TFUNCTION_H
#define TFUNCTION_H
struct functionT
{
  gsl_vector *para;
  gsl_vector *para_limit_left;
  gsl_vector *para_limit_right;
  gsl_vector_int *para_fit_yn;
  double *Q;
  double *I;
  double *dI;
  double *SIGMA;
  int *listLastPoints;
  int currentM;
  int currentFirstPoint;
  int currentLastPoint;
  int currentPoint;
  bool polyYN;
  int polyFunction;
  bool beforeFit;
  bool afterFit;
  bool beforeIter;
  bool afterIter;
  double Int1;
  double Int2;
  double Int3;
  int currentInt;
  int prec;
  char *tableName;
  char **tableColNames;
  int *tableColDestinations;
  gsl_matrix *mTable;
  int currentFunction;
};
#endif

/////////////////////////////////////////////////////////////////////////////////
//+++ Rounding of doubles to selected precition
/////////////////////////////////////////////////////////////////////////////////
double round2prec(double f, int prec)
{
  if(f==0.0) return 0.0;
double sign=1.0; if (f<0.0) sign=-1.0;
  f=fabs(f);
  if (f>1.0) prec--;
  int precData=(int)log10(f);
  f= rint( f / pow(10, precData) * pow(10,prec))/pow(10,prec)*pow(10, precData);
  return sign*f;
}

/////////////////////////////////////////////////////////////////////////////////
//+++ Global definition of parameters : new > 2019.06
/////////////////////////////////////////////////////////////////////////////////
static double A, Gamma, Beta, Bgd;
static bool initNow;

#define Para ((struct functionT *) ParaM)->para

#define XXX   ((struct functionT *) ParaM)->Q
#define YYY   ((struct functionT *) ParaM)->I
#define dYYY  ((struct functionT *) ParaM)->dI
#define RESO  ((struct functionT *) ParaM)->SIGMA
#define prec  ((struct functionT *) ParaM)->prec

#define listLastPoints    ((struct functionT *) ParaM)->listLastPoints
#define currentM          ((struct functionT *) ParaM)->currentM
#define currentFirstPoint ((struct functionT *) ParaM)->currentFirstPoint
#define currentLastPoint  ((struct functionT *) ParaM)->currentLastPoint
#define currentPoint      ((struct functionT *) ParaM)->currentPoint
#define currentInt        ((struct functionT *) ParaM)->currentInt
#define currentFunction   ((struct functionT *) ParaM)->currentFunction

#define beforeFit ((struct functionT *) ParaM)->beforeFit
#define afterFit  ((struct functionT *) ParaM)->afterFit
#define beforeIter ((struct functionT *) ParaM)->beforeIter
#define afterIter ((struct functionT *) ParaM)->afterIter
#define polyYN    ((struct functionT *) ParaM)->polyYN
#define Int1    ((struct functionT *) ParaM)->Int1
#define Int2    ((struct functionT *) ParaM)->Int2
#define Int3    ((struct functionT *) ParaM)->Int3

/////////////////////////////////////////////////////////////////////////////////
//+++ Function: read Parameters : new > 2019.06
/////////////////////////////////////////////////////////////////////////////////
void readParameters(void * ParaM)
{
  if (beforeFit || afterFit || beforeIter) initNow=true; else initNow=false;
  
  A=gsl_vector_get(Para, 0);
  Gamma=gsl_vector_get(Para, 1);
  Beta=gsl_vector_get(Para, 2);
  Bgd=gsl_vector_get(Para, 3);
}

/////////////////////////////////////////////////////////////////////////////////
//+++ Function: save Parameters : new > 2019.06
/////////////////////////////////////////////////////////////////////////////////
void saveParameters(void * ParaM)
{
  if (!beforeFit && !afterFit && !beforeIter && !afterIter) return;
  gsl_vector_set(Para,0,round2prec(A,prec));
  gsl_vector_set(Para,1,round2prec(Gamma,prec));
  gsl_vector_set(Para,2,round2prec(Beta,prec));
  gsl_vector_set(Para,3,round2prec(Bgd,prec));
}

/////////////////////////////////////////////////////////////////////////////////
//+++ Function returns Name of Function
/////////////////////////////////////////////////////////////////////////////////
extern "C" MY_EXPORT char *name()
{
  return "stretchedExp";
}

/////////////////////////////////////////////////////////////////////////////////
//+++ Function return List of names of Parameters 
/////////////////////////////////////////////////////////////////////////////////
extern "C" MY_EXPORT char *parameters()
{
  return "A,Gamma,Beta,Bgd,t";
}

/////////////////////////////////////////////////////////////////////////////////
//+++ Initial value of Parameters  
/////////////////////////////////////////////////////////////////////////////////
extern "C" MY_EXPORT char *init_parameters()
{
  return "1,1,1,0";
}

/////////////////////////////////////////////////////////////////////////////////
//+++ Initial 'Adjustibility' of Parameters  
/////////////////////////////////////////////////////////////////////////////////
extern "C" MY_EXPORT char *adjust_parameters()
{
  return "1,1,1,0";
}

/////////////////////////////////////////////////////////////////////////////////
//+++ Number of parameters
/////////////////////////////////////////////////////////////////////////////////
extern "C" MY_EXPORT char *paraNumber()
{
  return "4";
}

/////////////////////////////////////////////////////////////////////////////////
//+++ Description of Function and parameters
/////////////////////////////////////////////////////////////////////////////////
extern "C" MY_EXPORT char *listComments()
{
  char  *list;
  list="<html><head><meta name= \"qrichtext \" content= \"1 \" /></head><body style= \"font-size:8pt;font-family:MS Shell Dlg 2 \"><p dir= \"ltr \">Stretched exponential function:</p><p dir= \"ltr \">.. math:: I(t) =A e^{ - ( t Gamma) ^ Beta }</p><p dir= \"ltr \"></p></body></html>"
			",,Amplitude,,Relaxation rate in units 1/[unit t],,Stretched exponent,,Background";
 return list;
}

/////////////////////////////////////////////////////////////////////////////////
//+++ Included Functions
/////////////////////////////////////////////////////////////////////////////////



/////////////////////////////////////////////////////////////////////////////////
//+++ Fit Function
/////////////////////////////////////////////////////////////////////////////////
extern "C" MY_EXPORT double functionSANS(double key, void * ParaM)
{
//+++++++++++++++++++++++++++++++++++++++++++++
	 double t=key;
	 double It;
	 readParameters(ParaM);
//+++++++++++++++++++++++++++++++++++++++++++++
	 gsl_set_error_handler_off();
//+++++++++++++++++++++++++++++++++++++++++++++

// --->  It = f ( t, {P1, ...} );
It = A*exp( - pow (t*Gamma,Beta) ) + Bgd;

//+++++++++++++++++++++++++++++++++++++++++++++
	 saveParameters(ParaM);
	 return round2prec(It, prec);
//+++++++++++++++++++++++++++++++++++++++++++++
}
